/* -*-c-*-
 *
 * Low-level stuff for all Rijndael block sizes
 *
 * (c) 2001 Straylight/Edgeware
 */

/*----- Licensing notice --------------------------------------------------*
 *
 * This file is part of Catacomb.
 *
 * Catacomb is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Library General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * Catacomb is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with Catacomb; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA.
 */

/*----- Header files ------------------------------------------------------*/

#include "config.h"

#include <assert.h>
#include <stdio.h>

#include <mLib/bits.h>

#include "blkc.h"
#include "dispatch.h"
#include "gcipher.h"
#include "rijndael.h"
#include "rijndael-base.h"

/*----- Global variables --------------------------------------------------*/

const octet rijndael_keysz[] = { KSZ_RANGE, RIJNDAEL_KEYSZ, 4, 32, 4 };

/*----- Main code ---------------------------------------------------------*/

/* --- @rijndael_setup@ --- *
 *
 * Arguments:	@rijndael_ctx *k@ = pointer to context to initialize
 *		@unsigned nb@ = number of words in the block
 *		@const void *buf@ = pointer to buffer of key material
 *		@size_t sz@ = size of the key material
 *
 * Returns:	---
 *
 * Use:		Low-level key-scheduling.
 */

static void simple_setup(rijndael_ctx *k, unsigned nb,
			 const void *buf, unsigned nk)
{
  unsigned nr = k->nr, nw;
  unsigned i, j, jj;
  const octet *p;
  uint32 ww;

  /* --- Fetch the first key words out --- */

  p = buf;
  for (i = 0; i < nk; i++) {
    k->w[i] = LOAD32_B(p);
    p += 4;
  }

  /* --- Expand this material to fill the rest of the table --- */

  nw = (nr + 1) * nb;
  ww = k->w[i - 1];
  p = RCON;
  for (; i < nw; i++) {
    uint32 w = k->w[i - nk];
    if (i % nk == 0) {
      ww = ROL32(ww, 8);
      w ^= SUB(S, ww, ww, ww, ww) ^ (*p++ << 24);
    } else if (nk > 6 && i % nk == 4)
      w ^= SUB(S, ww, ww, ww, ww);
    else
      w ^= ww;
    k->w[i] = ww = w;
  }

  /* --- Make the decryption keys --- */

  j = nw; i = 0;

  j -= nb; jj = 0;
  for (; i < nb; i++)
    k->wi[i] = k->w[j + jj++];

  for (; i < nw - nb; i += nb) {
    j -= nb;
    for (jj = 0; jj < nb; jj++) {
      uint32 w = k->w[j + jj];
      k->wi[i + jj] = MIX(U, w, w, w, w);
    }
  }

  j -= nb; jj = 0;
  for (; i < nw; i++)
    k->wi[i] = k->w[j + jj++];
}

CPU_DISPATCH(static, EMPTY, void, setup,
	     (rijndael_ctx *k, unsigned nb, const void *buf, unsigned nk),
	     (k, nb, buf, nk), pick_setup, simple_setup)

#if CPUFAM_X86 || CPUFAM_AMD64
extern setup__functype rijndael_setup_x86ish_aesni;
#endif
#if CPUFAM_ARMEL && HAVE_AS_ARMV8_CRYPTO
extern setup__functype rijndael_setup_arm_crypto;
#endif
#if CPUFAM_ARM64
extern setup__functype rijndael_setup_arm64_crypto;
#endif

static setup__functype *pick_setup(void)
{
#if CPUFAM_X86 || CPUFAM_AMD64
  DISPATCH_PICK_COND(rijndael_setup, rijndael_setup_x86ish_aesni,
		     cpu_feature_p(CPUFEAT_X86_AESNI));
#endif
#if CPUFAM_ARMEL && HAVE_AS_ARMV8_CRYPTO
  DISPATCH_PICK_COND(rijndael_setup, rijndael_setup_arm_crypto,
		     cpu_feature_p(CPUFEAT_ARM_AES));
#endif
#if CPUFAM_ARM64
  DISPATCH_PICK_COND(rijndael_setup, rijndael_setup_arm64_crypto,
		     cpu_feature_p(CPUFEAT_ARM_AES));
#endif
  DISPATCH_PICK_FALLBACK(rijndael_setup, simple_setup);
}

void rijndael_setup(rijndael_ctx *k, unsigned nb, const void *buf, size_t sz)
{
  unsigned nk, nr;

  /* --- Sort out the key size --- */

  KSZ_ASSERT(rijndael, sz);
  nk = sz / 4;

  /* --- Select the number of rounds --- */

  nr = (nk > nb ? nk : nb) + 6;
  if (nr < 10)
    nr = 10;
  k->nr = nr;

  /* --- Do the main setup --- */

  setup(k, nb, buf, nk);
}

/*----- That's all, folks -------------------------------------------------*/
