/* -*-c-*-
 *
 * Implementation of the SHA-1 hash function
 *
 * (c) 1999 Straylight/Edgeware
 */

/*----- Licensing notice --------------------------------------------------*
 *
 * This file is part of Catacomb.
 *
 * Catacomb is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Library General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * Catacomb is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with Catacomb; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA.
 */

/*----- Header files ------------------------------------------------------*/

#include <mLib/bits.h>

#include "ghash.h"
#include "ghash-def.h"
#include "hash.h"
#include "sha.h"

/*----- Main code ---------------------------------------------------------*/

/* --- @sha_compress@ --- *
 *
 * Arguments:	@sha_ctx *ctx@ = pointer to context block
 *		@const void *sbuf@ = pointer to buffer of appropriate size
 *
 * Returns:	---
 *
 * Use:		SHA-1 compression function.
 */

void sha_compress(sha_ctx *ctx, const void *sbuf)
{
  uint32 a, b, c, d, e;
  uint32 m[16];
  const octet *p;
  int i;

  a  = ctx->a; b  = ctx->b; c  = ctx->c; d  = ctx->d; e  = ctx->e;
  for (p = sbuf, i = 0; i < 16; i++, p += 4) m[i] = LOAD32(p);

  /* --- Definitions for round functions --- */

#define F(x, y, z) (((x) & (y)) | (~(x) & (z)))
#define G(x, y, z) ((x) ^ (y) ^ (z))
#define H(x, y, z) (((x) & (y)) | ((x) & (z)) | ((y) & (z)))

#define T(v, w, x, y, z, i, f, k) do {					\
  z = ROL32(v, 5) + f(w, x, y) + z + m[i] + k;				\
  w = ROR32(w, 2);							\
} while (0)

#define FF(v, w, x, y, z, i) T(v, w, x, y, z, i, F, 0x5a827999)
#define GG(v, w, x, y, z, i) T(v, w, x, y, z, i, G, 0x6ed9eba1)
#define HH(v, w, x, y, z, i) T(v, w, x, y, z, i, H, 0x8f1bbcdc)
#define II(v, w, x, y, z, i) T(v, w, x, y, z, i, G, 0xca62c1d6)

  /* --- Message scheduling --- */

#define M(i, i3, i8, i14) do {						\
  uint32 t = m[i] ^ m[i3] ^ m[i8] ^ m[i14];				\
  m[i] = ROL32(t, 1);							\
} while (0)

  /* --- The main compression function --- */

  FF(a, b, c, d, e,  0);	M( 0, 13,  8,  2);
  FF(e, a, b, c, d,  1);	M( 1, 14,  9,  3);
  FF(d, e, a, b, c,  2);	M( 2, 15, 10,  4);
  FF(c, d, e, a, b,  3);	M( 3,  0, 11,  5);
  FF(b, c, d, e, a,  4);	M( 4,  1, 12,  6);
  FF(a, b, c, d, e,  5);	M( 5,  2, 13,  7);
  FF(e, a, b, c, d,  6);	M( 6,  3, 14,  8);
  FF(d, e, a, b, c,  7);	M( 7,  4, 15,  9);
  FF(c, d, e, a, b,  8);	M( 8,  5,  0, 10);
  FF(b, c, d, e, a,  9);	M( 9,  6,  1, 11);
  FF(a, b, c, d, e, 10);	M(10,  7,  2, 12);
  FF(e, a, b, c, d, 11);	M(11,  8,  3, 13);
  FF(d, e, a, b, c, 12);	M(12,  9,  4, 14);
  FF(c, d, e, a, b, 13);	M(13, 10,  5, 15);
  FF(b, c, d, e, a, 14);	M(14, 11,  6,  0);
  FF(a, b, c, d, e, 15);	M(15, 12,  7,  1);
  FF(e, a, b, c, d,  0);	M( 0, 13,  8,  2);
  FF(d, e, a, b, c,  1);	M( 1, 14,  9,  3);
  FF(c, d, e, a, b,  2);	M( 2, 15, 10,  4);
  FF(b, c, d, e, a,  3);	M( 3,  0, 11,  5);
  GG(a, b, c, d, e,  4);	M( 4,  1, 12,  6);
  GG(e, a, b, c, d,  5);	M( 5,  2, 13,  7);
  GG(d, e, a, b, c,  6);	M( 6,  3, 14,  8);
  GG(c, d, e, a, b,  7);	M( 7,  4, 15,  9);
  GG(b, c, d, e, a,  8);	M( 8,  5,  0, 10);
  GG(a, b, c, d, e,  9);	M( 9,  6,  1, 11);
  GG(e, a, b, c, d, 10);	M(10,  7,  2, 12);
  GG(d, e, a, b, c, 11);	M(11,  8,  3, 13);
  GG(c, d, e, a, b, 12);	M(12,  9,  4, 14);
  GG(b, c, d, e, a, 13);	M(13, 10,  5, 15);
  GG(a, b, c, d, e, 14);	M(14, 11,  6,  0);
  GG(e, a, b, c, d, 15);	M(15, 12,  7,  1);
  GG(d, e, a, b, c,  0);	M( 0, 13,  8,  2);
  GG(c, d, e, a, b,  1);	M( 1, 14,  9,  3);
  GG(b, c, d, e, a,  2);	M( 2, 15, 10,  4);
  GG(a, b, c, d, e,  3);	M( 3,  0, 11,  5);
  GG(e, a, b, c, d,  4);	M( 4,  1, 12,  6);
  GG(d, e, a, b, c,  5);	M( 5,  2, 13,  7);
  GG(c, d, e, a, b,  6);	M( 6,  3, 14,  8);
  GG(b, c, d, e, a,  7);	M( 7,  4, 15,  9);
  HH(a, b, c, d, e,  8);	M( 8,  5,  0, 10);
  HH(e, a, b, c, d,  9);	M( 9,  6,  1, 11);
  HH(d, e, a, b, c, 10);	M(10,  7,  2, 12);
  HH(c, d, e, a, b, 11);	M(11,  8,  3, 13);
  HH(b, c, d, e, a, 12);	M(12,  9,  4, 14);
  HH(a, b, c, d, e, 13);	M(13, 10,  5, 15);
  HH(e, a, b, c, d, 14);	M(14, 11,  6,  0);
  HH(d, e, a, b, c, 15);	M(15, 12,  7,  1);
  HH(c, d, e, a, b,  0);	M( 0, 13,  8,  2);
  HH(b, c, d, e, a,  1);	M( 1, 14,  9,  3);
  HH(a, b, c, d, e,  2);	M( 2, 15, 10,  4);
  HH(e, a, b, c, d,  3);	M( 3,  0, 11,  5);
  HH(d, e, a, b, c,  4);	M( 4,  1, 12,  6);
  HH(c, d, e, a, b,  5);	M( 5,  2, 13,  7);
  HH(b, c, d, e, a,  6);	M( 6,  3, 14,  8);
  HH(a, b, c, d, e,  7);	M( 7,  4, 15,  9);
  HH(e, a, b, c, d,  8);	M( 8,  5,  0, 10);
  HH(d, e, a, b, c,  9);	M( 9,  6,  1, 11);
  HH(c, d, e, a, b, 10);	M(10,  7,  2, 12);
  HH(b, c, d, e, a, 11);	M(11,  8,  3, 13);
  II(a, b, c, d, e, 12);	M(12,  9,  4, 14);
  II(e, a, b, c, d, 13);	M(13, 10,  5, 15);
  II(d, e, a, b, c, 14);	M(14, 11,  6,  0);
  II(c, d, e, a, b, 15);	M(15, 12,  7,  1);
  II(b, c, d, e, a,  0);
  II(a, b, c, d, e,  1);
  II(e, a, b, c, d,  2);
  II(d, e, a, b, c,  3);
  II(c, d, e, a, b,  4);
  II(b, c, d, e, a,  5);
  II(a, b, c, d, e,  6);
  II(e, a, b, c, d,  7);
  II(d, e, a, b, c,  8);
  II(c, d, e, a, b,  9);
  II(b, c, d, e, a, 10);
  II(a, b, c, d, e, 11);
  II(e, a, b, c, d, 12);
  II(d, e, a, b, c, 13);
  II(c, d, e, a, b, 14);
  II(b, c, d, e, a, 15);

  /* --- Update the chaining variables --- */

  ctx->a += a; ctx->b += b; ctx->c += c; ctx->d += d; ctx->e += e;
}

/* --- @sha_init@ --- *
 *
 * Arguments:	@sha_ctx *ctx@ = pointer to context block to initialize
 *
 * Returns:	---
 *
 * Use:		Initializes a context block ready for hashing.
 */

void sha_init(sha_ctx *ctx)
{
  ctx->a = 0x67452301;
  ctx->b = 0xefcdab89;
  ctx->c = 0x98badcfe;
  ctx->d = 0x10325476;
  ctx->e = 0xc3d2e1f0;
  ctx->off = 0;
  ctx->nl = ctx->nh = 0;
}

/* --- @sha_set@ --- *
 *
 * Arguments:	@sha_ctx *ctx@ = pointer to context block
 *		@const void *buf@ = pointer to state buffer
 *		@unsigned long count@ = current count of bytes processed
 *
 * Returns:	---
 *
 * Use:		Initializes a context block from a given state.  This is
 *		useful in cases where the initial hash state is meant to be
 *		secret, e.g., for NMAC and HMAC support.
 */

void sha_set(sha_ctx *ctx, const void *buf, unsigned long count)
{
  const octet *p = buf;
  ctx->a = LOAD32(p +  0);
  ctx->b = LOAD32(p +  4);
  ctx->c = LOAD32(p +  8);
  ctx->d = LOAD32(p + 12);
  ctx->e = LOAD32(p + 16);
  ctx->off = 0;
  ctx->nl = U32(count);
  ctx->nh = U32(((count & ~(unsigned long)MASK32) >> 16) >> 16);
}

/* --- @sha_hash@ --- *
 *
 * Arguments:	@sha_ctx *ctx@ = pointer to context block
 *		@const void *buf@ = buffer of data to hash
 *		@size_t sz@ = size of buffer to hash
 *
 * Returns:	---
 *
 * Use:		Hashes a buffer of data.  The buffer may be of any size and
 *		alignment.
 */

void sha_hash(sha_ctx *ctx, const void *buf, size_t sz)
{
  HASH_BUFFER(SHA, sha, ctx, buf, sz);
}

/* --- @sha_done@ --- *
 *
 * Arguments:	@sha_ctx *ctx@ = pointer to context block
 *		@void *hash@ = pointer to output buffer
 *
 * Returns:	---
 *
 * Use:		Returns the hash of the data read so far.
 */

void sha_done(sha_ctx *ctx, void *hash)
{
  octet *p = hash;
  HASH_PAD(SHA, sha, ctx, 0x80, 0, 8);
  STORE32(ctx->buf + SHA_BUFSZ - 8, (ctx->nl >> 29) | (ctx->nh << 3));
  STORE32(ctx->buf + SHA_BUFSZ - 4, ctx->nl << 3);
  sha_compress(ctx, ctx->buf);
  STORE32(p +  0, ctx->a);
  STORE32(p +  4, ctx->b);
  STORE32(p +  8, ctx->c);
  STORE32(p + 12, ctx->d);
  STORE32(p + 16, ctx->e);
}

/* --- @sha_state@ --- *
 *
 * Arguments:	@sha_ctx *ctx@ = pointer to context
 *		@void *state@ = pointer to buffer for current state
 *
 * Returns:	Number of bytes written to the hash function so far.
 *
 * Use:		Returns the current state of the hash function such that
 *		it can be passed to @sha_set@.
 */

unsigned long sha_state(sha_ctx *ctx, void *state)
{
  octet *p = state;
  STORE32(p +  0, ctx->a);
  STORE32(p +  4, ctx->b);
  STORE32(p +  8, ctx->c);
  STORE32(p + 12, ctx->d);
  STORE32(p + 16, ctx->e);
  return (ctx->nl | ((ctx->nh << 16) << 16));
}

/* --- Generic interface --- */

GHASH_DEF(SHA, sha)

/* --- Test code --- */

HASH_TEST(SHA, sha)

/*----- That's all, folks -------------------------------------------------*/
